@@ -19,7 +19,7 @@ of them as functions that are called on a glyph string.
 
 from fontTools.ttLib import TTFont
 from collections import OrderedDict
-from fontTools.feaLib.ast import ValueRecord
+from fontTools.feaLib.ast import ValueRecord as feaLibValueRecord
 from itertools import chain
 from bidict import bidict
 
@@ -41,15 +41,19 @@ class FontFeatures:
         self.symbols[category] += 1
         return f'{category}{self.symbols[category]}'
 
-    def addRoutine(self, r):
-        """Add a Routine to the preamble.
+    def referenceRoutine(self, r):
+        """Store a routine and return a reference to it.
 
         Args:
             r: A :py:class:`Routine` object.
         """
         assert isinstance(r, Routine)
-        self.routines.append(r)
+
+        if r not in self.routines:
+            self.routines.append(r)
         r.parent = self
+        r.usecount = r.usecount + 1
+        return RoutineReference(routine=r)
 
     def getNamedClassFor(self, glyphs, name):
         """Find and optionally stores a named class of glyphs
@@ -74,12 +78,14 @@ class FontFeatures:
 
         Args:
             name: The feature name.
-            rs: A sequence of :py:class:`Routine` objects.
+            rs: A sequence of :py:class:`Routine` or :py:class:`RoutineReference` objects.
         """
         if not name in self.features:
             self.features[name] = []
         for r in rs:
             r.parent = self
+            if isinstance(r, Routine):
+                r = self.referenceRoutine(r)
             self.features[name].append(r)
 
     def allRoutines(self):
@@ -161,25 +167,44 @@ class FontFeatures:
             if k.languages:
                 for l in k.languages:
                     add_language(l)
-        for feat in self.features.values():
-            for thing in feat:
-                if hasattr(thing, "languages") and thing.languages:
-                    for l in thing.languages:
-                        add_language(l)
 
-        if count > 0 and not "DFLT" in scripts:
-            scripts["DFLT"] = []
-            scripts.move_to_end("DFLT", last=False)
-        if count > 0 and not "dflt" in scripts["DFLT"]:
-            scripts["DFLT"].insert(0, "dflt")
+        # if count > 0 and not "DFLT" in scripts:
+        #     scripts["DFLT"] = []
+        #     scripts.move_to_end("DFLT", last=False)
+        # if count > 0 and not "dflt" in scripts["DFLT"]:
+        #     scripts["DFLT"].insert(0, "dflt")
 
         self.scripts_and_languages = scripts
 
+    def hasScriptSupport(self, script):
+        self.hoist_languages()
+        return script in self.scripts_and_languages
+
+    def resolveAllRoutines(self):
+        for routine in self.routines:
+            for r in routine.rules:
+                if not isinstance(r, Chaining):
+                    continue
+                for lookuplistIx in range(len(r.lookups)):
+                    r.lookups[lookuplistIx] = self.ensureLookupsAreReferences(r.lookups[lookuplistIx])
+
+    def ensureLookupsAreReferences(self, lookuplist):
+        rv = []
+        if not lookuplist:
+            return None
+        for r in lookuplist:
+            if isinstance(r, RoutineReference):
+                r.resolve(self)
+            else:
+                r = self.referenceRoutine(r)
+            assert r.routine in self.routines
+            rv.append(r)
+        return rv
 
 class Routine:
     """Represent a Routine (similar to OT Lookup).
 
-    A routine is a set of rules, sometimes but always with an explicit name.
+    A routine is a set of rules, sometimes but not always with an explicit name.
     It can apply to a set of language/script pairs.
     """
     def __init__(
@@ -191,8 +216,11 @@ class Routine:
         languages=None,
         parent=None,
         flags=0,
+        markFilteringSet=None,
+        markAttachmentSet=None,
     ):
         self.name = name
+        self.usecount = 0
         self.usedin = set()
         if rules:
             self.rules = rules
@@ -201,9 +229,11 @@ class Routine:
         self.address = address
         self.comments = []
         self.inlined = inlined
-        self.languages = languages
+        self.languages = languages or []
         self.parent = parent
         self.flags = flags
+        self.markFilteringSet = markFilteringSet
+        self.markAttachmentSet = markAttachmentSet
 
     def addRule(self, rule):
         """Adds a rule to a Routine.
@@ -230,7 +260,79 @@ class Routine:
         """Returns the names of all of the glyphs involved in this Routine."""
         return set.union(*[r.involved_glyphs for r in self.rules])
 
+    @property
+    def stage(self):
+        for r in self.rules:
+            if isinstance(r, Substitution):
+                return "sub"
+            if isinstance(r, Positioning):
+                return "pos"
+            if isinstance(r, Attachment):
+                return "pos"
+            if isinstance(r, Chaining):
+                return r.stage
+
     from .feaLib.Routine import asFea, asFeaAST, feaPreamble
+    from .shaperLib.Routine import apply_to_buffer
+    from .xmlLib.Routine import toXML, fromXML
+
+class ExtensionRoutine(Routine):
+    def __init__(self, **kwargs):
+        if "routines" in kwargs:
+            self.routines = kwargs["routines"]
+            del kwargs["routines"]
+        super().__init__(**kwargs)
+
+    def apply_to_buffer(self, buf, stage=None, feature=None):
+        for r in self.routines:
+            r.apply_to_buffer(buf, stage, feature)
+
+    def asFeaAST(self):
+        import fontTools.feaLib.ast as feaast
+        f = feaast.Block()
+        for r in self.routines:
+            f.statements.append(r.asFeaAST())
+        return f
+
+    @property
+    def stage(self):
+        return self.routines[0].stage
+
+    @property
+    def rules(self):
+        return list(chain(*[routine.rules for routine in self.routines]))
+
+    @rules.setter
+    def rules(self, foo):
+        pass
+
+class RoutineReference:
+    def __init__(self, name=None, routine=None):
+        self.routine = routine
+        if self.routine:
+            self.name = routine.name
+        if name:
+            self.name = name
+
+    def resolve(self, ff):
+        if not self.routine:
+            for r in ff.routines:
+                if r.name == self.name:
+                    self.routine = r
+                    break
+            if not self.routine:
+                raise ValueError("Could not resolve routine")
+        if not self.name:
+            self.name = self.routine.name
+
+    @property
+    def stage(self):
+        if not self.routine:
+            raise ValueError("Routine reference not resolved")
+        return self.routine.stage
+
+    from .xmlLib.RoutineReference import fromXML, toXML
+    from .feaLib.RoutineReference import asFea, asFeaAST, feaPreamble
 
 
 class Rule:
@@ -242,6 +344,8 @@ class Rule:
         """Computes any text that needs to go in the feature file header."""
         return []
 
+    from .shaperLib.Rule import would_apply_at_position, pre_post_context_matches
+    from .xmlLib.Rule import fromXML, toXML, _makeglyphslots, _slotArray
 
 class Substitution(Rule):
     """Represents a Substitution rule.
@@ -288,6 +392,7 @@ class Substitution(Rule):
         self.languages = languages
         self.flags = flags
         self.reverse = reverse
+        self.stage = "sub"
 
     @property
     def involved_glyphs(self):
@@ -298,6 +403,8 @@ class Substitution(Rule):
         return i | o | b | a
 
     from .feaLib.Substitution import asFeaAST
+    from .shaperLib.Substitution import shaper_inputs, _do_apply
+    from .xmlLib.Substitution import _toXML, fromXML
 
 
 class Chaining(Rule):
@@ -336,7 +443,15 @@ class Chaining(Rule):
         self.languages = languages
         self.flags = flags
 
-    from .feaLib.Chaining import asFeaAST, feaPreamble
+    @property
+    def stage(self):
+        for l in self.lookups:
+            if not l:
+                continue
+            for aLookup in l:
+                if not aLookup:
+                    continue
+                return aLookup.stage
 
     @property
     def involved_glyphs(self):
@@ -345,6 +460,68 @@ class Chaining(Rule):
         a = set(chain.from_iterable(self.postcontext))
         return i | b | a
 
+    from .feaLib.Chaining import asFeaAST, feaPreamble
+    from .shaperLib.Chaining import shaper_inputs, _do_apply
+    from .xmlLib.Chaining import _toXML, fromXML
+
+
+class ValueRecord(feaLibValueRecord):
+    """An extension of the fontTools.feaLib ValueRecord representation to support
+    variable font item variation stores."""
+
+    # Actually until we have a binary compiler, we're just going to cheat and store
+    # a value record for each master.
+    def set_value_for_master(self, vf, master, vr):
+        """Set a master-specific value record.
+
+        Args:
+            vf: a babelfont.VariableFont instance
+            master: name of the master to set
+            vr: a ``ValueRecord`` specific to this master.
+        """
+        if not hasattr(self, "masters"):
+            self.masters = {}
+            # Clone this value record to all masters
+            for mastername in vf.masters.keys():
+                self.masters[mastername] = ValueRecord(
+                    xPlacement=vr.xPlacement,
+                    yPlacement=vr.yPlacement,
+                    xAdvance=vr.xAdvance,
+                    yAdvance=vr.yAdvance)
+        else:
+            self.masters[master] = vr
+
+    def get_value_for_master(self, vf, master):
+        """Get a master-specific value record.
+
+        Args:
+            vf: a ``babelfont.VariableFont`` instance
+            master: name of the master to get
+        """
+        if not hasattr(self, "masters"):
+            return self
+        return self.masters[master]
+
+    def get_value_for_location(self, vf, location):
+        """Get an interpolated value record for a particular location.
+
+        Args:
+            vf: a ``babelfont.VariableFont`` instance
+            location: A dictionary mapping axis names to user space coordinates.
+        """
+        if not hasattr(self, "masters"):
+            return self
+        locs = {k: (v.xPlacement, v.yPlacement, v.xAdvance, v.yAdvance) for k,v in self.masters.items()}
+        values = vf.interpolate_tuples(
+            locs, location
+        )
+        xPlacement, yPlacement, xAdvance, yAdvance = values
+        return ValueRecord(
+            xPlacement=xPlacement,
+            yPlacement=yPlacement,
+            xAdvance=xAdvance,
+            yAdvance=yAdvance
+        )
 
 class Positioning(Rule):
     """Represents a Positioning rule.
@@ -378,6 +555,7 @@ class Positioning(Rule):
         self.address = address
         self.languages = languages
         self.flags = flags
+        self.stage = "pos"
 
     @property
     def involved_glyphs(self):
@@ -387,6 +565,8 @@ class Positioning(Rule):
         return i | b | a
 
     from .feaLib.Positioning import asFeaAST
+    from .shaperLib.Positioning import shaper_inputs, _do_apply
+    from .xmlLib.Positioning import _toXML, fromXML
 
 
 class Attachment(Rule):
@@ -414,12 +594,15 @@ class Attachment(Rule):
         self.flags = flags
         self.address = address
         self.font = font
+        self.stage = "pos"
 
     @property
     def is_cursive(self):
         return self.base_name == "cursive_entry" or self.base_name == "entry" # XXX
 
     from .feaLib.Attachment import asFeaAST, feaPreamble
+    from .shaperLib.Attachment import shaper_inputs, _do_apply, would_apply_at_position
+    from .xmlLib.Attachment import _toXML, fromXML
 
     @property
     def involved_glyphs(self):
