@@ -64,12 +64,7 @@ def asFeaAST(self):
         for base in self.baseslist:
             statementtype = feaast.MarkBasePosStatement
             if self.font:
-                if "GDEF" not in self.font:
-                    warnings.warn("# No GDEF table; can't distinguish marks and bases. mkmk won't work.")
-                elif base[0][0] not in self.font["GDEF"].table.GlyphClassDef.classDefs:
-                    warnings.warn("# Glyph %s not found in attachment; skipping" % base[0][0])
-                    continue
-                elif categorize_glyph(self.font,base[0][0])[0] == "mark":
+                if categorize_glyph(self.font,base[0][0])[0] == "mark":
                     statementtype = feaast.MarkMarkPosStatement
             b.statements.append(
                 statementtype(
